from dune.perftool import Restriction

from dune.perftool.options import get_option

from dune.perftool.pdelab.argument import name_coefficientcontainer

from dune.perftool.generation import (class_member,
                                      constructor_block,
                                      domain,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      include_file,
                                      initializer_list,
                                      temporary_variable,
                                      valuearg
                                      )

from dune.perftool.pdelab.localoperator import (name_domain_field,
                                                lop_template_range_field,
                                                )
from dune.perftool.pdelab.quadrature import estimate_quadrature_order

from pytools import Record

import numpy


class AMatrix(Record):
    def __init__(self, a_matrix, m, n):
        Record.__init__(self,
                        a_matrix=a_matrix,
                        m=m,
                        n=n,
                        )


@class_member("operator")
def define_alignment(name):
    alignment = get_option("sumfact_alignment")
    return "enum {{ {} = {} }};".format(name, str(alignment))


def name_alignment():
    name = "alignment"
    define_alignment(name)
    return name


def quadrature_points_per_direction():
    # TODO use quadrature order from dune.perftool.pdelab.quadrature
    # as soon as we have a generic implementation

    # Quadrature order
    q = estimate_quadrature_order()

    # Quadrature points in per direction
    nb_qp = q // 2 + 1

    return nb_qp


@class_member("operator")
def define_number_of_quadrature_points_per_direction(name):
    number_qp = quadrature_points_per_direction()
    return "enum {{ {} = {} }};".format(name, str(number_qp))


def name_number_of_quadrature_points_per_direction():
    name = "m"
    define_number_of_quadrature_points_per_direction(name)
    return name


def polynomial_degree():
    form = get_global_context_value("formdata").preprocessed_form
    return form.coefficients()[0].ufl_element()._degree


def basis_functions_per_direction():
    return polynomial_degree() + 1


@class_member("operator")
def define_number_of_basis_functions_per_direction(name):
    number_basis = basis_functions_per_direction()
    return "enum {{ {} = {} }};".format(name, str(number_basis))


def name_number_of_basis_functions_per_direction():
    name = "n"
    define_number_of_basis_functions_per_direction(name)
    return name


@class_member("operator")
def define_oned_quadrature_weights(name):
    range_field = lop_template_range_field()
    number_qp = name_number_of_quadrature_points_per_direction()
    return "{} {}[{}];".format(range_field, name, number_qp)


def name_oned_quadrature_weights():
    name = "qw"
    define_oned_quadrature_weights(name)
    return name


@class_member("operator")
def define_oned_quadrature_points(name):
    range_field = lop_template_range_field()
    number_qp = name_number_of_quadrature_points_per_direction()
    return "{} {}[{}];".format(range_field, name, number_qp)


def name_oned_quadrature_points():
    name = "qp"
    define_oned_quadrature_points(name)
    return name


@class_member("operator")
def typedef_polynomials(name):
    range_field = lop_template_range_field()
    domain_field = name_domain_field()
    degree = polynomial_degree()

    include_file("dune/pdelab/finiteelementmap/qkdg.hh", filetag="operatorfile")

    # TODO: make switching polynomials possible
    # return "using {} = Dune::QkStuff::GaussLobattoLagrangePolynomials<{}, {}, {}>;".format(name,
    #                                                                                        domain_field,
    #                                                                                        range_field,
    #                                                                                        degree)
    return "using {} = Dune::QkStuff::EquidistantLagrangePolynomials<{}, {}, {}>;".format(name,
                                                                                          domain_field,
                                                                                          range_field,
                                                                                          degree)


def type_polynomials():
    name = "Polynomials1D"
    typedef_polynomials(name)
    return name


@class_member("operator")
def define_polynomials(name):
    polynomials_type = type_polynomials()
    return "{} {};".format(polynomials_type, name)


def name_polynomials():
    name = "poly"
    define_polynomials(name)
    return name


@constructor_block("operator")
def sort_quadrature_points_weights():
    range_field = lop_template_range_field()
    domain_field = name_domain_field()
    number_qp = name_number_of_quadrature_points_per_direction()
    qp = name_oned_quadrature_points()
    qw = name_oned_quadrature_weights()
    include_file("dune/perftool/sumfact/onedquadrature.hh", filetag="operatorfile")
    return "onedQuadraturePointsWeights<{}, {}, {}>({}, {});".format(range_field, domain_field, number_qp, qp, qw)


@constructor_block("operator")
def construct_theta(name):
    # Make sure that the quadrature points are sorted
    sort_quadrature_points_weights()

    m = name_number_of_quadrature_points_per_direction()
    n = name_number_of_basis_functions_per_direction()
    polynomials = name_polynomials()
    qp = name_oned_quadrature_points()

    return ["for (std::size_t i=0; i<{}; i++){{".format(m),
            "  for (std::size_t j=0; j<{}; j++){{".format(n),
            "    {}.colmajoraccess(i,j) = {}.p(j,{}[i]);".format(name, polynomials, qp),
            "  }",
            "}"]


@class_member("operator")
def typedef_theta(name):
    include_file("dune/perftool/sumfact/alignedmatvec.hh", filetag="operatorfile")
    alignment = name_alignment()
    range_field = lop_template_range_field()
    return "using {} = AlignedMatrix<{}, {}>;".format(name, range_field, alignment)


def type_theta():
    name = "AMatrix"
    typedef_theta(name)
    return name


@class_member("operator")
def define_theta(name):
    theta_type = type_theta()
    number_qp = quadrature_points_per_direction()
    number_basis = basis_functions_per_direction()
    globalarg(name, shape=(number_basis, number_qp), dtype=numpy.float32, dim_tags="f,f")
    initializer_list(name, [str(number_qp), str(number_basis)], classtag="operator")
    construct_theta(name)
    return "{} {};".format(theta_type, name)


def name_theta():
    name = "Theta"
    define_theta(name)
    return name
